/*-----------------------------------------------------------------------------
  Project:  QRE
  File:     Graphicscontrol.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Revision: 1.6 $
 *----------------------------------------------------------------------------*/
#include <nn/os.h>
#include <nn/gx.h>

#include "GraphicsControl.h"

using namespace nn::gx;

GraphicsControl::GraphicsControl()
{
}

GraphicsControl::~GraphicsControl()
{
}

int GraphicsControl::InitializeGraphics(nn::fnd::ExpHeap* heap)
{
    s_AppHeap = heap;
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap->Allocate(0x800000));
    s_RenderSystem.Initialize(s_HeapForGx, 0x800000);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.f, 0.f, 0.f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.f, 0.f, 0.f, 1.0f);

    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    return 0;
}

void GraphicsControl::FinalizeGraphics(void)
{
    s_RenderSystem.Finalize();
    s_AppHeap->Free(reinterpret_cast<void*>(s_HeapForGx));
}

void GraphicsControl::ClearDisplay(const s32 display)
{
    s_RenderSystem.SetRenderTarget(display);
    s_RenderSystem.Clear();
}

void GraphicsControl::SwapBuffer(const s32 display, const bool wait)
{
    s_RenderSystem.SwapBuffers();
    if (wait)
        s_RenderSystem.WaitVsync(display);
}

void GraphicsControl::DrawImage(u8* textureDataPtr, u32 width, u32 height)
{
    static GLuint textureId = 0;
    if (textureId != 0)
        s_RenderSystem.DeleteTexture(textureId);
    GLenum target = GL_TEXTURE_2D;
    GLenum internalFormat = GL_RGB_NATIVE_DMP;
    GLenum format = GL_RGB_NATIVE_DMP;
    GLenum type = GL_UNSIGNED_BYTE;
    u32 textureWidth = width;
    u32 textureHeight = height;
    u8* textureDataBuffer = GetTextureDataFromRawData(width, height,
                                                      textureDataPtr,
                                                      textureWidth,
                                                      textureHeight);
    s_RenderSystem.GenerateTexture(target, internalFormat, textureWidth, textureHeight,
                                   format, type, textureDataBuffer, textureId);
    s_AppHeap->Free(reinterpret_cast<void*>(textureDataBuffer));

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    f32 windowPositionX = (DISPLAY0_HEIGHT / 2) - (width / 2);
    f32 windowPositionY = 0.0f;
    if (height < DISPLAY0_WIDTH)
        windowPositionY = (DISPLAY0_WIDTH / 2) - (height / 2);
    f32 rectangleWidth = static_cast<f32>(width);
    f32 rectangleHeight = static_cast<f32>(height);
    s_RenderSystem.FillTexturedRectangle(textureId,
                                         windowPositionX, windowPositionY,
                                         rectangleWidth, rectangleHeight,
                                         static_cast<f32>(width),
                                         static_cast<f32>(height),
                                         static_cast<f32>(textureWidth),
                                         static_cast<f32>(textureHeight));
}

u8* GraphicsControl::GetTextureDataFromRawData(const u32& width,
                                               const u32& height,
                                               u8* rawDataBuffer,
                                               u32& textureWidth,
                                               u32& textureHeight)
{
    textureWidth = GetTextureLength(width);
    textureHeight = GetTextureLength(height);
    u32 textureSize = textureWidth * textureHeight * 3;
    u8* textureGLDataBuffer = reinterpret_cast<u8*>(s_AppHeap->Allocate(textureSize));
    u8* dstPtr = textureGLDataBuffer;

    u32 y = 0;
    for (; y < height; y++) {
#if 0
        u8* srcPtr = rawDataBuffer + width * y * 3;
#else
        // FLIP対応
        // ConvertGLTextureToNative()にてFLIPしている。
        u8* srcPtr = rawDataBuffer + width * (height - y - 1) * 3;
#endif
        u32 x = 0;
        for (; x < width; x++) {
            dstPtr[0] = srcPtr[2];
            dstPtr[1] = srcPtr[1];
            dstPtr[2] = srcPtr[0];
            dstPtr += 3;
            srcPtr += 3;
        }
        for (; x < textureWidth; x++) {
            dstPtr[0] = dstPtr[1] = dstPtr[2] = 0x00;
            dstPtr += 3;
        }
    }
    for (; y < textureHeight; y++) {
        for (u32 x = 0; x < textureWidth; x++) {
            dstPtr[0] = dstPtr[1] = dstPtr[2] = 0x00;
            dstPtr += 3;
        }
    }

    u8* textureDataBuffer = reinterpret_cast<u8*>(s_AppHeap->Allocate(textureSize));
    GLenum format = GL_RGB_NATIVE_DMP;
    bool result = demo::ConvertGLTextureToNative(format, textureWidth, textureHeight,
                                                 textureGLDataBuffer, textureDataBuffer);
    if (!result) {
        NN_LOG("  Convert to GL_RGB_NATIVE_DMP failed.\n");
        NN_ASSERT(0);
    }
    s_AppHeap->Free(reinterpret_cast<void*>(textureGLDataBuffer));

    return textureDataBuffer;
}

u32 GraphicsControl::GetTextureLength(const u32& imageLength)
{
    u32 textureLength = 8;
    // 8, 16, 32, 64, 128, 256, 512, 1024
    for (u32 i = 0; i < 7; i++) {
        if (imageLength < textureLength)
            return textureLength;

        textureLength *= 2;
    }
    return 1024;
}
